@@ -4,6 +4,7 @@ import os
 from abc import ABC, abstractmethod
 import json
 import itertools
+from functools import wraps
 
 import hashlib
 from urllib.parse import urlparse
@@ -11,6 +12,11 @@ from urllib.parse import parse_qs
 
 from pathlib import PurePosixPath
 
+try:
+    from collections.abc import Sequence
+except ImportError:
+    from collections import Sequence
+
 import google_auth_httplib2
 from googleapiclient.discovery import build
 from google_auth_oauthlib.flow import InstalledAppFlow
@@ -22,7 +28,11 @@ from googleapiclient.http import MediaFileUpload
 from googleapiclient.http import MediaUploadProgress
 from googleapiclient.http import MediaDownloadProgress
 
-from google.auth.exceptions import RefreshError
+from expiringdict import ExpiringDict
+from datetime import datetime, timedelta
+
+from drivelib.errors import GoogleDriveAPIError
+from drivelib.errors import BackendError
 
 import logging
 logger = logging.getLogger('drivelib')
@@ -103,17 +113,69 @@ class ResumableMediaUploadProgress(MediaUploadProgress):
                                 self.resumable_uri
                             )
 
+def needs_id(f):
+    @wraps(f)
+    def wrapper(self, *args, **kwargs):
+        if not self.id:
+            raise FileNotFoundError
+        return f(self, *args, **kwargs)
+
+    return wrapper
+
+def autorefresh(f):
+    @wraps(f)
+    def wrapper(self, *args, **kwargs):
+        if self.last_refreshed + self.refresh_after <= datetime.now():
+            self.refresh()
+        return f(self, *args, **kwargs)
+    return wrapper
+
+class _DriveParents(Sequence):
+    """This object provides sequence-like access to the logical ancestors
+    of a path."""
+    def __init__(self, startpoint: DriveItem):
+        self._startpoint = startpoint
+        self._parents = dict()
+
+    def __getitem__(self, idx):
+        if idx in self._parents and self._parents[idx] is not None:
+            return self._parents[idx]
+        if idx == 0:
+            self._parents[idx] = self._startpoint.parent
+        else:
+            try:
+                self._parents[idx] = self[idx - 1].parent
+            except AttributeError:
+                raise IndexError
+
+        if self._parents[idx] is None:
+            raise IndexError
+
+        return self._parents[idx]
+
+    def __len__(self):
+        l = 0
+        while True:
+            try:
+                self[l]
+            except IndexError:
+                return l
+            else:
+                l += 1
+
 class DriveItem(ABC):
     #TODO: metadata as dict
     # Filename not as attribute but as key
     # OR: filename as property method
 
-    def __init__(self, drive, parent_ids, name, id_, spaces):
+    def __init__(self, drive, parent_ids, name, id_, spaces, refresh_after=timedelta(minutes=30)):
         self.drive = drive
         self.name = name # TODO: property. Setter => rename
         self.id = id_ # TODO: property
         self.parent_ids = parent_ids
         self.spaces = spaces
+        self.last_refreshed = datetime.now()
+        self.refresh_after = refresh_after
 
     def __eq__(self, other):
         return self.id == other.id
@@ -123,11 +185,22 @@ class DriveItem(ABC):
         return hash(self.id)
 
     @property
+    @autorefresh
     def parent(self):
-        if self.parent_ids:
-            return self.drive.item_by_id(self.parent_ids[0])
-        else:
-            return None
+        if not hasattr(self, '_parent'):
+            if self.parent_ids:
+                self._parent = self.drive.item_by_id(self.parent_ids[0])
+            else:
+                self._parent = None
+        return self._parent
+
+    @property
+    def parents(self):
+        "An immutable sequence providing access to the logical ancestors of the path"
+        return _DriveParents(self)
+
+    def resolve(self) -> str:
+        return '/'.join((x.name if x != self.drive else '' for x in reversed(self.parents)))+'/'+self.name
 
     def rename(self, target, ignore_existing=False):
         splitpath = target.rsplit('/', 1)
@@ -142,8 +215,11 @@ class DriveItem(ABC):
 
         self.move(parent, new_name, ignore_existing=ignore_existing)
 
+    @needs_id
     def move(self, new_dest, new_name=None, ignore_existing=False):
         new_name = new_name or self.name
+        if new_dest == self.parent and new_name == self.name:
+            return
         if not ignore_existing:
             try:
                 child = new_dest.child(new_name)
@@ -154,18 +230,25 @@ class DriveItem(ABC):
                 raise FileExistsError("Filename already exists multiple times: {name}".format(name=new_name))
 
         #TODO: Don't use exception for flow control here. Maybe implement exists()
-        result = self.drive.service.files().update(
-                                fileId=self.id,
-                                body={"name": new_name},
-                                addParents=new_dest.id,
-                                removeParents=self.parent.id,
-                                fields='name, parents',
-                                ).execute()
+        try:
+            result = self.drive.service.files().update(
+                                    fileId=self.id,
+                                    body={"name": new_name},
+                                    addParents=new_dest.id,
+                                    removeParents=self.parent.id,
+                                    fields='name, parents',
+                                    ).execute()
+        except HttpError as err:
+            raise GoogleDriveAPIError.from_http_error(err)
         self.name = result['name']
         self.parent_ids = result.get('parents', [])
         
+    @needs_id
     def remove(self):
-        self.drive.service.files().delete(fileId=self.id).execute()
+        try:
+            self.drive.service.files().delete(fileId=self.id).execute()
+        except HttpError as err:
+            raise GoogleDriveAPIError.from_http_error(err)
         self.id = None
 
     def trash(self):
@@ -174,36 +257,97 @@ class DriveItem(ABC):
         except:
             raise HttpError("Could not trash file")
 
+    @needs_id
     def meta_set(self, metadata: dict):
-        result = self.drive.service.files().update(
-                                fileId=self.id,
-                                body=metadata,
-                                fields=','.join(metadata.keys()),
-                                ).execute()
+        try:
+            result = self.drive.service.files().update(
+                                    fileId=self.id,
+                                    body=metadata,
+                                    fields=','.join(metadata.keys()),
+                                    ).execute()
+        except HttpError as err:
+            raise GoogleDriveAPIError.from_http_error(err)
         #TODO update local array
 
+    @needs_id
     def meta_get(self, fields: str) -> dict:
         #TODO cache metadata
-        return self.drive.service.files().get(fileId=self.id, fields=fields).execute()
+        try:
+            return self.drive.service.files().get(fileId=self.id, fields=fields).execute()
+        except HttpError as err:
+            raise GoogleDriveAPIError.from_http_error(err)
 
+    @needs_id
     def refresh(self):
-        result = self.drive.service.files().get(
-                                fileId=self.id,
-                                fields=self.drive.default_fields
-                            ).execute()
+        try:
+            result = self.drive.service.files().get(
+                                    fileId=self.id,
+                                    fields=self.drive.default_fields
+                                ).execute()
+        except HttpError as err:
+            raise GoogleDriveAPIError.from_http_error(err)
         self.name = result['name']
         self.parent_ids = result['parents']
+        self.last_refreshed = datetime.now()
+
+    @needs_id
+    def create_shortcut(self, name, parent=None) -> DriveShortcut:
+        if parent:
+            parent_id = parent.id
+        else:
+            try:
+                parent_id = self.parent.id
+            except AttributeError:
+                parent_id = 'root'
+                
+        shortcut_metadata = {
+            'name': name,
+            'parents': [parent_id],
+            'mimeType': 'application/vnd.google-apps.shortcut',
+            'shortcutDetails': {
+                'targetId': self.id
+            }
+        }
+        try:
+            result = self.drive.service.files().create(
+                                            body=shortcut_metadata,
+                                            fields=self.drive.default_fields).execute()
+        except HttpError as err:
+            raise GoogleDriveAPIError.from_http_error(err)
+        return self._reply_to_object(result)
+
+    def _reply_to_object(self, reply) -> DriveItem:
+        if reply['mimeType'] == 'application/vnd.google-apps.folder':
+            new_item = DriveFolder(self.drive, reply.get('parents', []), reply['name'], reply['id'], spaces=",".join(reply['spaces']))
+        elif reply['mimeType'] == 'application/vnd.google-apps.shortcut':
+            new_item = DriveShortcut(self.drive, reply.get('parents', []), reply['name'], reply['id'], reply['shortcutDetails']['targetId'], spaces=",".join(reply['spaces']))
+        else:
+            new_item = DriveFile(self.drive, reply.get('parents', []), reply['name'], reply['id'], spaces=",".join(reply['spaces']))
+        self.drive._id_cache[new_item.id] = new_item
+        return new_item
+
 
     @abstractmethod
-    def isfolder(self):
-        pass
+    def isfolder(self) -> bool:
+        raise NotImplementedError
+
+    @abstractmethod
+    def isshortcut(self) -> bool:
+        raise NotImplementedError
 
 class DriveFolder(DriveItem):
 
-    def isfolder(self):
+    def isfolder(self) -> bool:
         return True  
+
+    def isshortcut(self) -> bool:
+        return False
     
-    def child(self, name):
+    def child(self, name) -> DriveItem:
+        child = self._get_from_name_cache(name)
+        if child:
+            logger.debug("Found {} in name cache".format(name))
+            return child
         gen = self.children(name=name, pageSize=2)
         child = next(gen, None)
         if child is None:
@@ -212,9 +356,22 @@ class DriveFolder(DriveItem):
         if next_child is not None:
             duplicates = itertools.chain((child, next_child), gen)
             raise AmbiguousPathError("Two or more files {name}".format(name=name), duplicates=duplicates)
+        self._add_to_name_cache(child)
         return child
-        
-    def children(self, name=None, folders=True, files=True, trashed=False, pageSize=100, orderBy=None):
+
+    @needs_id
+    def _add_to_name_cache(self, item: DriveItem):
+        self.drive._name_cache['/'.join((self.id, item.name))] = item
+
+    @needs_id
+    def _delete_from_name_cache(self, item: DriveItem):
+        del self.drive._name_cache['/'.join((self.id, item.name))]
+
+    def _get_from_name_cache(self, name: str):
+        return self.drive._name_cache.get('/'.join(((self.id, name))), None)
+
+    @needs_id
+    def children(self, name=None, folders=True, files=True, trashed=False, pageSize=100, orderBy=None, skip=0) -> Iterator(DriveItem):
         query = "'{this}' in parents".format(this=self.id)
 
         if name:
@@ -232,9 +389,10 @@ class DriveFolder(DriveItem):
         else:
             query += " and trashed = false"
 
-        return self.drive.items_by_query(query, pageSize=pageSize, orderBy=orderBy, spaces=self.spaces)
+        return self.drive.items_by_query(query, pageSize=pageSize, orderBy=orderBy, spaces=self.spaces, skip=skip)
 
-    def mkdir(self, name, ignore_existing=False):
+    @needs_id
+    def mkdir(self, name, ignore_existing=False) -> DriveFolder:
         if not ignore_existing:
             try:
                 file_ = self.child(name)
@@ -254,10 +412,14 @@ class DriveFolder(DriveItem):
             'mimeType': 'application/vnd.google-apps.folder',
             'parents': [self.id]
         }
-        result = self.drive.service.files().create(body=file_metadata, fields=self.drive.default_fields).execute()
+        try:
+            result = self.drive.service.files().create(body=file_metadata, fields=self.drive.default_fields).execute()
+        except HttpError as err:
+            raise GoogleDriveAPIError.from_http_error(err)
         return self._reply_to_object(result)
         
-    def new_file(self, filename, ignore_existing=False):
+    @needs_id
+    def new_file(self, filename, ignore_existing=False) -> DriveFile:
         if not ignore_existing:
             try:
                 self.child(filename)
@@ -310,27 +472,23 @@ class DriveFolder(DriveItem):
         else:
             return False
         
-    def _reply_to_object(self, reply):
-        if reply['mimeType'] == 'application/vnd.google-apps.folder':
-            return DriveFolder(self.drive, reply.get('parents', []), reply['name'], reply['id'], spaces=",".join(reply['spaces']))
-        else:
-            return DriveFile(self.drive, reply.get('parents', []), reply['name'], reply['id'], spaces=",".join(reply['spaces']))
-
 class DriveFile(DriveItem):  
 
     def isfolder(self):
         return False  
 
+    def isshortcut(self):
+        return False
+
     def __init__(self, drive, parent_ids, filename, file_id=None, spaces='drive', resumable_uri=None):
         super().__init__(drive, parent_ids, filename, file_id, spaces)
         self.resumable_uri = resumable_uri
         
+    @needs_id
     def download(self, local_file, chunksize=None, progress_handler=None):
         if not chunksize:
             chunksize = defaultChunksize
         #TODO: Accept Path objects for local_file
-        if not self.id:
-            raise FileNotFoundError
 
         range_md5 = hashlib.md5()
         try:
@@ -342,15 +500,11 @@ class DriveFile(DriveItem):
             local_file_size = 0
         
 
-        remote_file_size = int(self.drive.service.files().\
-                            get(fileId=self.id, fields="size").\
-                            execute()['size'])
-        
         download_url = "https://www.googleapis.com/drive/v3/files/{fileid}?alt=media".\
                                 format(fileid=self.id)
         
         with open(local_file, 'ab') as fh:
-            while local_file_size < remote_file_size:
+            while local_file_size < self.size:
                 download_range = "bytes={}-{}".\
                     format(local_file_size, local_file_size+chunksize-1)
                     
@@ -364,9 +518,9 @@ class DriveFile(DriveItem):
                         local_file_size+=int(resp['content-length'])
                         range_md5.update(content)
                         if progress_handler:
-                            progress_handler(MediaDownloadProgress(local_file_size, remote_file_size))
+                            progress_handler(MediaDownloadProgress(local_file_size, self.size))
                 else:
-                    raise HttpError(resp, content)
+                    raise GoogleDriveAPIError.from_reply(resp, content)
         if range_md5.hexdigest() != self.md5sum:
             os.remove(local_file)
             raise CheckSumError("Checksum mismatch. Need to repeat download.")
@@ -413,9 +567,37 @@ class DriveFile(DriveItem):
             'name': self.name, 
             'parents': self.parent_ids
         }
-        result = self.drive.service.files().create(body=file_metadata, fields=self.drive.default_fields).execute()
+        try:
+            result = self.drive.service.files().create(body=file_metadata, fields=self.drive.default_fields).execute()
+        except HttpError as err:
+            raise GoogleDriveAPIError.from_http_error(err)
         self.id = result['id']
         self.name = result['name']
+
+    def copy(self, dest=None, new_name=None, ignore_existing=False):
+        if not dest:
+            dest = self.parent
+        if not new_name:
+            new_name = self.name
+        if not ignore_existing:
+            try:
+                dest.child(new_name)
+                raise FileExistsError
+            except FileNotFoundError:
+                pass
+        
+        try:
+            result = self.drive.service.files().copy(
+                                fileId=self.id,
+                                body={
+                                    "name": new_name,
+                                    "parents": [dest.id]
+                                },
+                                fields=self.drive.default_fields,
+                                ).execute()
+        except HttpError as err:
+            raise GoogleDriveAPIError.from_http_error(err)
+        return dest._reply_to_object(result)
        
     @property
     def md5sum(self):
@@ -429,6 +611,28 @@ class DriveFile(DriveItem):
             self._size = int(self.meta_get("size")["size"])
         return self._size
 
+class DriveShortcut(DriveItem):
+    def __init__(self, drive, parent_ids, filename, file_id, target_id, spaces='drive'):
+        super().__init__(drive, parent_ids, filename, file_id, spaces)
+        self.target = self.drive.item_by_id(target_id)
+
+    def isshortcut(self) -> bool:
+        return True
+
+    def isfolder(self) -> bool:
+        return False
+
+    def meta_get(self, fields: str) -> dict:
+        return self.target.meta_get(fields)
+
+    def meta_set(self, metadata: dict):
+        return self.target.meta_set(metadata)
+
+    def resolve(self) -> str:
+        return self.target.resolve()
+
+    def __getattr__(self, name):
+        return getattr(self.target, name)
 
 class ResumableUploadRequest:
     # TODO: actually implement interface for http_request
@@ -458,7 +662,7 @@ class ResumableUploadRequest:
             api_url = "https://www.googleapis.com/upload/drive/v3/files?uploadType=resumable" 
             status, resp = self.service._http.request(api_url, method='POST', headers={'Content-Type':'application/json; charset=UTF-8'}, body=json.dumps(self.body)) 
             if status['status'] != '200':
-                raise HttpError(status, resp)
+                raise GoogleDriveAPIError.from_reply(status, resp)
             self._resumable_uri = status['location']
         return self._resumable_uri
         
@@ -475,7 +679,7 @@ class ResumableUploadRequest:
             
             if status['status'] not in ('200', '308'):
                 #Should 404 result in a FileNotFound error?
-                raise HttpError(status, resp)
+                raise GoogleDriveAPIError.from_reply(status, resp)
 
             self._range_md5 = hashlib.md5()
             def file_in_chunks(start_byte: int, end_byte: int, chunksize: int = 4*1024**2):
@@ -530,12 +734,12 @@ class ResumableUploadRequest:
                     if e.resp.status == 404:
                         raise FileNotFoundError("File was successfully uploaded but since has been deleted")
                     else:
-                        raise
+                        raise GoogleDriveAPIError.from_http_error(e)
                 logger.debug("Remote MD5 (0-%d): %s", self.resumable_progress, remote_md5)
                 if remote_md5 != self._range_md5.hexdigest():
                     raise CheckSumError("Final checksum mismatch. Need to repeat upload.")
         else:
-            raise HttpError(status, resp)
+            raise GoogleDriveAPIError.from_reply(status, resp)
             
         return ResumableMediaUploadProgress(self.resumable_progress, self.media_body.size(), self.resumable_uri), resp
 
@@ -575,7 +779,7 @@ class GoogleDrive(DriveFolder):
             except (AssertionError, KeyError):
                 raise InvalidUrlError(url)
 
-    def __init__(self, creds, autorefresh=True):
+    def __init__(self, creds, autorefresh=True, caching=1000):
         try:
             self.creds = Credentials.from_json(creds)
         except TypeError:
@@ -592,18 +796,18 @@ class GoogleDrive(DriveFolder):
 
         self._service = build('drive', 'v3', http=http)
 
+        self._id_cache = ExpiringDict(max_len=caching, max_age_seconds=float('inf'))
+        self._name_cache = ExpiringDict(max_len=caching, max_age_seconds=60)
+
         self.id = None
         self.drive = self
-        self.default_fields = 'id, name, mimeType, parents, spaces'
+        self.default_fields = 'id, name, mimeType, parents, spaces, shortcutDetails'
         root_folder = self.item_by_id("root")
 
         super().__init__(self, root_folder.parent_ids, root_folder.name, root_folder.id, root_folder.spaces)
 
         if 'https://www.googleapis.com/auth/drive.appdata' in self.creds.scopes:
             self.appdata = self.item_by_id("appDataFolder")
-        
-        #self.caching = caching
-        #TODO: Add caching ability
 
     @property
     def service(self):
@@ -612,27 +816,51 @@ class GoogleDrive(DriveFolder):
     def json_creds(self):
         return Credentials.to_json(self.creds)
 
-    def items_by_query(self, query, pageSize=100, orderBy=None, spaces='drive'):
+    def items_by_query(self, query, pageSize=100, orderBy=None, spaces='drive', skip=0) -> DriveItem:
+        pageSize = min(1000, max(pageSize, skip))
         result = {'nextPageToken': ''}
         while "nextPageToken" in result:
-            result = self.service.files().list(
-                    pageSize=pageSize,
-                    spaces=spaces,
-                    fields="nextPageToken, files({})".format(self.default_fields),
-                    q=query,
-                    pageToken=result['nextPageToken'],
-                    orderBy=orderBy,
-                ).execute()
+            try:
+                result = self.service.files().list(
+                        pageSize=pageSize,
+                        spaces=spaces,
+                        fields="nextPageToken, files({})".format(self.default_fields),
+                        q=query,
+                        pageToken=result['nextPageToken'],
+                        orderBy=orderBy,
+                    ).execute()
+            except HttpError as err:
+                raise GoogleDriveAPIError.from_http_error(err)
             items = result.get('files', [])
 
-            for file_ in items:
+            for i, file_ in enumerate(items):
+                if skip > i:
+                    continue
+                if skip == i:
+                    pageSize = 100
                 yield self._reply_to_object(file_)
 
-    def item_by_id(self, id_):
+    def item_by_id(self, id_) -> DriveItem:
         if hasattr(self, 'id') and id_ == self.id:
             return self
-        result = self.service.files().get(
-                                fileId=id_,
-                                fields=self.default_fields
-                            ).execute()
+
+        if id_ in self._id_cache:
+            logger.debug("Found {} in id cache".format((id_)))
+            return self._id_cache[id_]
+
+        try:
+            result = self.service.files().get(
+                                    fileId=id_,
+                                    fields=self.default_fields
+                                ).execute()
+        except HttpError as err:
+            raise GoogleDriveAPIError.from_http_error(err)
         return self._reply_to_object(result)
+
+    def resolve(self) -> str:
+        return '/'
+    # def create_shortcut(self, *args, **kwargs):
+    #     try:
+    #         super().create_shortcut(*args, **kwargs)
+    #     except BackendError:
+    #         raise BackendError("Can't create shortcut to root")
